#!/usr/bin/env python

from pathlib import Path
import os
import fftwlib
from fftwlib import encodeMIPS
from fftwlib import bootSign
from fftwlib import ebootDecrypt
from fftwlib import encodeTextFiles
import pycdlib
import shutil
try:
    from cStringIO import StringIO as BytesIO
except ImportError:
    from io import BytesIO

def applyFFTASMPatches(bfin,bkl,fin,fou,fou2):
    patches,locations,location,specialWords=[],[],[],[]
    locFile,locType="",""
    offset,newPatch,desSection,defSection,badLocation=0,0,False,False,True
    with open(bfin+fin,encoding="utf8",errors="ignore") as file:
        for line_number,line in enumerate(file,start=1):
            line=line.strip()
            if line[0]=="#":
                continue
            else:
                if line.find("#")!=-1:
                    line=line.split("#")[0]
                if line[0:2]=="$-" and line.find(":")!=-1:
                    slines=line.split(":",1)
                    if "$-name"==slines[0]:
                        if location:
                            locations.append(location)
                            patches.append(locations)
                        location,locations,newPatch=[],[],1
                    elif "$-uuid"==slines[0] and newPatch==1:
                        uuid,newPatch=slines[1],2
                        for exclude in bkl:
                            if uuid==exclude:
                                print("   Excluding: "+uuid)
                                newPatch=-1
                                break
                        if newPatch==2:
                            locations.append(uuid)
                    elif "$-description"==slines[0]:
                        desSection,defSection,badLocation=True,False,True
                        continue
                    elif "$-overwrites"==slines[0]:
                        continue
                    elif "$-requires"==slines[0] and newPatch==2:
                        if slines[1]!="none":
                            if requirementsAreNotBefore(patches,slines[1]):
                                print("   Failed. Missing requirements for: "+str(locations[0]))
                                return True
                    elif "$-define"==slines[0] and newPatch>=2:
                        desSection,defSection,badLocation=False,True,True
                    elif "$-file"==slines[0] and newPatch>=2:
                        desSection,defSection,badLocation=False,False,True
                        locFile=slines[1]
                        newPatch=3
                    elif "$-type"==slines[0] and newPatch>=3:
                        desSection,defSection,badLocation=False,False,True
                        locType=slines[1]
                        newPatch=4
                    elif "$-offset"==slines[0] and newPatch==4:
                        desSection,defSection,badLocation=False,False,False
                        if location:
                            locations.append(location)
                            location=[]
                        if slines[1].find("&")!=-1:
                            lparts=slines[1].split("&")
                        else:
                            lparts=[slines[1]]
                        offset=getOffsetLocation(specialWords,lparts[0])
                        if locFile=="boot.bin" and locType=="ram":
                            location.append(offset-fftwlib.PSPramOffset)
                        elif locFile=="boot.bin" and locType=="file":
                            location.append(offset)
                        else:
                            inFile=Path(bfin+locType)
                            if inFile.is_file():
                                location.append(offset)
                                if len(lparts)==2:
                                    location.append("$File$"+bfin+locType+"$"+lparts[1])
                                else:
                                    location.append("$File$"+bfin+locType+"$")
                                if locFile=="boot.bin":
                                    location.append(fou)
                                else:
                                    location.append(fou2)
                            elif locType=="NULL_FILE" and len(lparts)==2:
                                location.append(offset)
                                location.append("$File$"+locType+"$"+lparts[1])
                                if locFile=="boot.bin":
                                    location.append(fou)
                                else:
                                    location.append(fou2)
                            elif locType=="relative" and len(lparts)==3:
                                location.append(offset)
                                location.append("$Relative$"+lparts[1]+"$"+lparts[2])
                                if locFile=="boot.bin":
                                    location.append(fou)
                                else:
                                    location.append(fou2)
                            else:
                                badLocation=True
                                print("   Bad location in "+fin+" at: "+str(line_number))
                    else:
                        if newPatch!=-1:
                            print("   Warning. Unidentified instruction: "+slines[0])
                        continue
                elif desSection:
                    continue
                elif defSection:
                    if "#" in line:
                        line=line.split("#")[0]
                        if not line:
                            continue
                    line=line.strip()
                    if line[0]=="%":
                        specialWords.append([line.split(",")[0],line.split(",")[1],-1,len(patches)])
                        specialWords.sort(reverse=True)
                elif not badLocation:
                    if "#" in line:
                        line=line.split("#")[0]
                        if not line or line.isspace():
                            continue
                    line=line.strip()
                    if  line[0]=="@":
                        specialWords.append([line.split(":")[0],f"{offset:#0{10}X}".lower(),-1,-1])
                        specialWords.sort(reverse=True)
                    elif ":" in line:
                        location.append((f"{offset:#0{10}X}".lower())[2:]+": "+line.split(":")[1])
                        specialWords.append([line.split(":")[0],f"{offset:#0{10}X}".lower(),
                                             len(locations),len(patches)])
                        specialWords.sort(reverse=True)
                        offset+=4
                    else:
                        location.append((f"{offset:#0{10}X}".lower())[2:]+": "+line)
                        offset+=4
    if location:
        locations.append(location)
        patches.append(locations)
    patches=replaceSpecialWords(patches,specialWords)
    writePatches(patches,fou)
    return False
def getOffsetLocation(specialWords,sOffset):
    for sword in specialWords:
        if sword[0]==sOffset:
            sOffset=sword[1]
            break
    return int(sOffset,16)
def readWords(patch):
    words=[]
    for line in patch:
        if "#" in line:
            line=line.split("#")[0]
        lparts=line.split(":")
        lparts[1]=lparts[1].strip()
        if "noop*" in lparts[1]:
            numNoops=int(lparts[1].split("*")[1])
            for i in range(0,numNoops):
                words.append(encodeMIPS.setEncode("noop",lparts[0]))
        elif "encodeText;" in lparts[1]:
            words.append(["text",
                          encodeTextFiles.encodeTextCore(lparts[1].split(";",1)[1],
                                                         False)[0][0]])
        elif "writeBits;" in lparts[1]:
            words.append(["bits",lparts[1].split(";",1)[1]])
        elif "writeHalf;" in lparts[1]:
            words.append(["half",lparts[1].split(";",1)[1]])
        else:
            words.append(encodeMIPS.setEncode(lparts[1],lparts[0]))
    return words
def replaceSpecialWords(patches,specialWords):
    newpatches,ipatches=[],0
    for locations in patches:
        newlocations,ilocations=[],0
        newlocations.append(locations[0])
        for location in locations[1:]:
            newlocation=[]
            newlocation.append(location[0])
            ilocations+=1
            for line in location[1:]:
                for sword in specialWords:
                    if sword[0] in line and \
                       ((sword[3]==-1) or
                        (sword[2]==-1 and sword[3]==ipatches) or
                        (sword[2]==ilocations and sword[3]==ipatches)):
                        line=line.replace(sword[0],sword[1])
                newlocation.append(line)
            newlocations.append(newlocation)
        newpatches.append(newlocations)
        ipatches+=1
    return newpatches
def writePatches(patches,fou):
    for locations in patches:
        locations.pop(0)
        for location in locations:
            if "$File$" in location[1]:
                if location[1].split("$")[2]=="NULL_FILE":
                    nullFileReplacement(location[0],location[1].split("$")[3],location[2])
                else:
                    binaryFileReplacement(location[0],location[1].split("$")[3],
                                          location[1].split("$")[2],location[2])
            elif "$Relative$" in location[1]:
                offset=location[0]
                if location[1].split("$")[2]=="INDEXED":
                    with open(location[2],"rb") as file:
                        file.seek(offset)
                        offset=int.from_bytes(file.read(4),"little",signed=False)
                offset+=int(location[1].split("$")[3],16)
                words=checkEndianess(readWords(location[3:]))
                with open(location[2],"r+b") as file:
                    file.seek(offset)
                    for word in words:
                        file.write(word)
            else:
                offset=location[0]
                words=checkEndianess(readWords(location[1:]))
                with open(fou,"r+b") as file:
                    file.seek(offset)
                    for word in words:
                        file.write(word)
    return
def checkEndianess(list):
    words=[]
    for word in list:
        try:
            if word<0:
                print("  Bad instruction: "+str(word))
            words.append(word.to_bytes(4,"little",signed=False))
        except:
            if word[0]=="text":
                for sword in word[1]:
                    words.append(sword.to_bytes(1,"big",signed=False))
            elif word[0]=="bits":
                words.append(int(word[1],2).to_bytes(1,"big",signed=False))
            elif word[0]=="half":
                words.append(int(word[1],16).to_bytes(2,"little",signed=False))
    return words
def binaryFileReplacement(offset,eparam,fin,fou):
    if eparam=="INDEXED":
        with open(fou,"rb") as file:
            file.seek(offset)
            offset=int.from_bytes(file.read(4),"little",signed=False)
    with open(fin,"rb") as infile:
        data=infile.read()
        with open(fou,"r+b") as file:
            file.seek(offset)
            file.write(data)
    return
def nullFileReplacement(offset,eparam,fou):
    if eparam=="INDEXED":
        with open(fou,"rb") as file:
            file.seek(offset)
            offset=int.from_bytes(file.read(4),"little",signed=False)
            isize=int.from_bytes(file.read(4),"little",signed=False)-offset
    else:
        isize=int(eparam)
    data=b"\x00"*isize
    with open(fou,"r+b") as file:
        file.seek(offset)
        file.write(data)
    return
def requirementsAreNotBefore(patches,requirements):
    if requirements.find("&")!=-1:
        requirements=requirements.split("&")
    else:
        requirements=[requirements]
    for requirement in requirements:
        reqFound=False
        for locations in patches:
            if requirement==locations[0]:
                reqFound=True
                break
        if not reqFound:
            return True
    return False
def extractFFTPack(inFile,outFile):
    print("Extracting fftpack.bin:")
    iso=pycdlib.PyCdlib()
    iso.open(inFile)
    extracted=BytesIO()
    iso.get_file_from_iso_fp(extracted,iso_path='/PSP_GAME/USRDIR/fftpack.bin')
    extracted.seek(0)
    with open(outFile,"wb") as file:
        shutil.copyfileobj(extracted,file,length=131072)
    iso.close()
    print(" Done.")
    return
def copyBOOT_Pack_ISO(inFile,outFile,inFile2,outFile2,inFile3,outFile3):
    print("Copying boot.bin:")
    copyFile(inFile,outFile)
    print(" Done.")
    print("Copying fftpack.bin:")
    copyFile(inFile2,outFile2)
    print(" Done.")
    print("Copying fft-wotl.iso:")
    copyFile(inFile3,outFile3)
    print(" Done.")
    return
def extractAndDecryptEBOOT(inFile,outFile):
    print("Extracting EBOOT.BIN:")
    outFolder=outFile[:outFile.rfind("/")+1]
    iso=pycdlib.PyCdlib()
    iso.open(inFile)
    extracted=BytesIO()
    iso.get_file_from_iso_fp(extracted,iso_path='/PSP_GAME/SYSDIR/EBOOT.BIN')
    extracted.seek(0)
    with open(outFolder+'EBOOT.BIN',"wb") as file:
        shutil.copyfileobj(extracted,file,length=131072)
    iso.close()
    print(" Done.")
    print("Decrypting EBOOT.BIN:")
    ebootDecrypt.decrypt(outFolder+'EBOOT.BIN',outFile,"../libkirk/")
    return
def signBOOT(inFile,outFolder):
    #For PBP's,to be done.
    return bootSign.sign(inFile,outFolder+"eboot-signed.bin",7,"../libkirk/")
def copyFile(fin,fou):
    with open(fou,"wb") as ofile:
        with open(fin,"rb") as file:
            ofile.write(file.read())
    return
def extractPackBOOTFiles(inFile,inFile2,outFolder,subfolders,files,offsets):
    os.makedirs(outFolder,exist_ok=True)
    for folder in subfolders:
        os.makedirs(outFolder+folder[0]+"/",exist_ok=True)
    idx=0
    for subfiles in files:
        jdx=0
        print("Extracting files: "+subfolders[idx][0])
        for subfile in subfiles:
            offset=offsets[idx][jdx][1]
            if offset==-1:
                jdx+=1
                continue
            print(" "+subfile)
            infile=inFile
            isize=1
            if offsets[idx][jdx][0]=="INDEXED":
                with open(infile,"rb") as file:
                    file.seek(offset)
                    offset=int.from_bytes(file.read(4),"little",signed=False)
                    isize=int.from_bytes(file.read(4),"little",signed=False)-offset
            elif offsets[idx][jdx][0]=="BOOT":
                infile=inFile2
                isize=offsets[idx][jdx][2]
            with open(infile,"rb") as file:
                file.seek(offset)
                data=file.read(isize)
            with open(outFolder+subfolders[idx][0]+"/"+subfile,"wb") as file:
                file.write(data)
            jdx+=1
        print(" Done.")
        idx+=1
    return
def coreFFTPatcher(patchFolder,blacklist,patches,outFile,outFile2,outFile3):
    if not Path(outFile).is_file() or not Path(outFile2).is_file():
        print("Missing output files. Aborting.")
        return
    bkl=[]
    print("Setting excluded hacks from:")
    print(" "+blacklist)
    if not Path(patchFolder+blacklist).is_file():
        print("  File not found.")
    else:
        with open(patchFolder+blacklist,encoding="utf8",errors="ignore") as file:
            for line_number,line in enumerate(file,start=1):
                line=line.strip()
                if line[0]=="#":
                    continue
                else:
                    if line.find("#")!=-1:
                        line=line.split("#")[0]
                    if line[0:2]=="$-" and line.find(":")!=-1:
                        slines=line.split(":",1)
                        if "$-uuid"==slines[0]:
                            bkl.append(slines[1])
                    print("  "+bkl[-1])
    print(" Done.")
    print("Applying patches in:")
    for dpatch in patches:
        if dpatch[1]=="disable":
            continue
        patch=dpatch[0]
        print(" "+patch+":")
        if not Path(patchFolder+patch).is_file():
            print("  File not found.")
            continue
        if applyFFTASMPatches(patchFolder,bkl,patch,outFile,outFile2):
            break
        print("  Applied.")
    print(" Done.")
    print("Patching iso:")
    iso=pycdlib.PyCdlib()
    iso.open(outFile3,"r+b")
    with open(outFile,"rb") as file:
        print(" /PSP_GAME/SYSDIR/EBOOT.BIN")
        size=os.path.getsize(outFile)+348
        iso.modify_file_in_place(file,size,iso_path="/PSP_GAME/SYSDIR/EBOOT.BIN")
        print("  "+"Patched.")
    with open(outFile2,"rb") as file:
        print(" /PSP_GAME/USRDIR/fftpack.bin")
        size=os.path.getsize(outFile2)
        iso.modify_file_in_place(file,size,iso_path="/PSP_GAME/USRDIR/fftpack.bin")
        print("  "+"Patched.")
    iso.close()
    print(" Done.")
    return
