#!/usr/bin/env python

import fftwlib
from fftwlib import decodeTextFiles
from fftwlib import eventMap
import multiprocessing as mp
from functools import partial
import struct
import math
import os

def decodeEVT(baseFolder,ouFile,inFile,patched,type):
    if ouFile[1]=="disable":
        return
    print("Decoding: "+ouFile[0])
    print(" CPU cores detected: "+str(mp.cpu_count()))
    baseFolder=baseFolder+ouFile[0]+"/"
    fin=baseFolder+inFile[0]
    fou=baseFolder+ouFile[0]
    offsets,offset=[],0
    limit=os.path.getsize(fin)
    while offset<limit:
        offsets.append(offset)
        offset+=fftwlib.EVTsize[type]
    decodePool=mp.Pool()
    decodeFunc=partial(decodeEVTCore,fin,patched,type)
    with open(fou+"-unt","w",encoding="utf8",errors="ignore",newline="") as file:
        for lines in decodePool.imap(decodeFunc,offsets):
            for line in lines:
                file.writelines(line)
    print(" Done.")
    return
def decodeEVTCore(fin,patched,type,offset):
    with open(fin,"rb") as file:
        file.seek(offset)
        data=file.read(fftwlib.EVTsize[type])
    textOffset=min(int.from_bytes(data[0:4],"little",signed=False),fftwlib.EVTsize[type])
    lines,tOffNew=decodeEVTInstructions(data[4:textOffset],
                                        textOffset,math.floor(offset/fftwlib.EVTsize[type]))
    lines.append("{event text}\n")
    if tOffNew<textOffset and textOffset>=len(data):
        lines.append("{zeroes("+str(fftwlib.EVTsize[type]-tOffNew)+","+str(tOffNew)+")}\n")
    else:
        lines.extend(decodeTextFiles.decodeTextCore(data[textOffset:],patched,textOffset))
    return lines
def decodeEVTInstructions(fdata,textOffset,nEVT):
    lines,offset=[],0
    lines.append("{event "+str(nEVT+1)+"}\n")
    if len(fdata)==0:
        return lines,textOffset
    while offset<len(fdata):
        data=fdata[offset]
        instSize=eventMap.EVTCommandSize[data]
        if offset+instSize > len(fdata):
            return lines,textOffset
        inst=eventMap.EVTVa2Na.get(data)
        lines.append(printStruct(inst[0],
                                 struct.unpack(inst[1],fdata[offset+1:offset+instSize]) if inst[1] else ""))
        if data==219 or data==227:
            textOffset=5
            if offset+1<len(fdata):
                nxdata=fdata[offset+1]
                if nxdata==219 or nxdata==227:
                    lines.append("EventEnd()\n")
                    textOffset+=1
            textOffset+=offset
            break
        offset+=instSize
    return lines,textOffset
def printStruct(base,line):
    if line=="":
        return base+"()\n"
    base+="("
    for field in line:
        base+=hex(field)+","
    return base[:-1]+")\n"
