#!/usr/bin/env python

from fftwlib import charMap
from fftwlib import textFileTools
import multiprocessing as mp
from functools import partial

Dictionary=charMap.getCharMap()
DictionaryP=charMap.getCharMap(True)

def decodeSectionedTextFile(fin,fou,patched):
    with open(fin,"rb") as file:
        fdata=file.read()
    header=fdata[0:0x80]
    bdata=fdata[0x80:]
    sections=[]
    for idx in range(0,32):
        coffset=header[idx*4:idx*4+4]
        coffset=int.from_bytes(coffset,"little")
        if idx < 31:
            iidx=idx+1
            noffset=header[iidx*4:iidx*4+4]
            noffset=int.from_bytes(noffset,"little")
        else:
            noffset=len(bdata)
        sections.append([coffset,noffset])
    lines,idx=[],1
    for section in sections:
        lines.append("{section "+f"{idx:02d}"+"}\n")
        idx+=1
        if section[0]!=section[1]:
            sdata=textFileTools.decompressText(bdata[section[0]:section[1]])
            slines=decodeTextCore(sdata,patched,section[0])
            for line in slines:
                lines.append(line)
    with open(fou,"w",encoding="utf8",errors="ignore",newline="") as file:
        for line in lines:
            file.writelines(line)
    return
def decodeTextFile(fin,fou,patched,size=0,cOff=0):
    offsets=[0]
    if size!=0:
        with open(fin,"rb") as file:
            fdata=file.read()
        for i in range(size,len(fdata),size):
            offsets.append(i)
        decodePool=mp.Pool()
        decodeFunc=partial(decodeText,fin,offsets,patched,cOff)
        indexs=list(range(0,len(offsets)))
        with open(fou,"w",encoding="utf8",errors="ignore",newline="") as file:
            for lines in decodePool.imap(decodeFunc,indexs):
                for line in lines:
                    file.writelines(line)
    else:
        lines=decodeText(fin,offsets,patched,cOff,0)
        with open(fou,"w",encoding="utf8",errors="ignore",newline="") as file:
            for line in lines:
                file.writelines(line)
    return
def decodeText(fin,offsets,patched,cOff,index):
    with open(fin,"rb") as file:
        file.seek(offsets[index])
        fdata=file.read(offsets[index+1]-offsets[index]) if index+1<len(offsets) else file.read()
    return decodeTextCore(textFileTools.decompressText(fdata,cOff),patched)
def decodeTextCore(fdata,patched,soffset=0):
    dictionary=charMap.getCharMap(patched)
    lines,line,offset=[],"",0
    if len(fdata)==1:
        data=fdata[offset]
        offset+=1
    elif len(fdata)>1:
        data=fdata[offset:offset+2]
        offset+=2
    else:
        return lines
    while data:
        twoFound=False
        aux=int.from_bytes(data,"big",signed=False)
        aux2=int.from_bytes(data[0:1],"big",signed=False)
        if aux!=0 and aux2>=0xd0 and aux2<0xf8:
            val=dictionary[aux-0xcf00]
            if "{INVALID}"!=val:
                [lines,line]=writeNewline(lines,line,val,offset,soffset)
                twoFound=True
        if not twoFound:
            aux=int.from_bytes(data[0:1],"big",signed=False)
            if patched:
                [lines,line]=writeNewline(lines,line,DictionaryP[aux],offset,soffset)
            else:
                [lines,line]=writeNewline(lines,line,Dictionary[aux],offset,soffset)
            if not (offset<len(fdata)):
                if len(data)>1:
                    aux=int.from_bytes(data[1:2],"big",signed=False)
                    if patched:
                        [lines,line]=writeNewline(lines,line,DictionaryP[aux],offset,soffset)
                    else:
                        [lines,line]=writeNewline(lines,line,Dictionary[aux],offset,soffset)
                data=[]
            else:
                pdata=fdata[offset:offset+1]
                offset+=1
                data=data[1:2]+pdata
        else:
            if not (offset+1<len(fdata)):
                data=fdata[offset:offset+1]
                offset+=1
            else:
                data=fdata[offset:offset+2]
                offset+=2
    [lines,line]=writeNewline(lines,line,"{Stop}",offset,soffset)
    return lines
def writeNewline(lines,line,value,offset,soffset):
    if line and len(line)>5 and line[-5:]==5*"0":
        if line[0:5]!=5*"0":
            lines.append(line+"\n")
            line=value
        elif value==len(value)*"0":
            line+=value
        else:
            offset=offset-len(line)+soffset
            lines.append("{zeroes("+str(len(line))+","+str(offset)+")}"+"\n")
            line=""
            if "Close}"==value[2:] or "{Newline}"==value:
                lines.append(value+"\n")
            elif "{Stop}"!=value:
                line=value
    elif "Close}"==value[2:]:
        if line:
            lines.append(line+"\n")
        lines.append(value+"\n")
        line=""
    elif "{Stop}"==value:
        if line:
            lines.append(line+"\n")
    else:
        line+=value
        if "{Newline}"==value:
            lines.append(line+"\n")
            line=""
    return [lines,line]
def decodeSNPLMesFile(subfile,subfile2,patched,size):
    decodeTextFile(subfile,subfile2+".txt",patched,size)
    textFileTools.extractEntriesNovels(subfile2+".txt",subfile2)
    return
def decodeSpellMesFile(subfile,subfile2,patched,size):
    decodeTextFile(subfile,subfile2+".txt",patched,size)
    textFileTools.extractEntriesSQ(subfile2+".txt",subfile2)
    return
def decodeOtherFile(subfile,subfile2,patched,size):
    if "LZW" in subfile:
        decodeSectionedTextFile(subfile,subfile2+"-unt",patched)
    else:
        decodeTextFile(subfile,subfile2+"-unt",patched,size,128)
    return
def decodeTextFiles(baseFolder,files,ifiles,patched):
    print("CPU cores detected: "+str(mp.cpu_count()))
    for idx in range(0,len(files)):
        if files[idx][1]=="disable":
            continue
        print("Decoding: "+files[idx][0])
        jdx,size=0,files[idx][3]
        for subfile in ifiles[idx]:
            if files[idx][2][jdx][0]==-1:
                jdx+=1
                continue
            print(" "+subfile)
            sfi=baseFolder+files[idx][0]+"/"+subfile
            sf2=baseFolder+files[idx][0]+"/"+files[idx][0]+"-"+subfile.split(".")[0]
            if idx==0:
                decodeSpellMesFile(sfi,sf2,patched,size)
            elif idx==1:
                decodeSNPLMesFile(sfi,sf2,patched,size)
            else:
                decodeOtherFile(sfi,sf2,patched,size)
            jdx+=1
        print(" Done.")
    return
