#!/usr/bin/env python

import fftwlib
from fftwlib import dataMIPS
import multiprocessing as mp
from functools import partial
import math

def disassemblyFFTbootMIPS(fin,fou,type="psp-fft-boot"):
    print("Disassembling: "+fin[max(fin.rfind("/")+1,0):])
    print(" CPU cores detected: "+str(mp.cpu_count()))
    with open(fin,mode="rb") as file:
        data=file.read()
    size=10240
    offsets,offset,limit=[],84,math.floor(len(data)/size)*size
    while offset<limit:
        offsets.append(offset)
        offset+=size
    decodePool,lines=mp.Pool(),[]
    decodeFunc=partial(disassemblyCore,fin,fftwlib.PSPramOffset,type,size)
    for words in decodePool.imap(decodeFunc,offsets):
        lines.extend(words)
    lines.extend(disassemblyCore(fin,fftwlib.PSPramOffset,"psp-fft-boot",0,limit))
    with open(fou,"w",encoding="utf8",errors="ignore",newline="") as file:
        for word in lines:
            file.writelines(word)
    print(" Done.")
    return
def disassemblyCore(fin,rOff=0,type="",size=0,offset=0):
    with open(fin,mode="rb") as file:
        file.seek(offset)
        data=file.read() if size==0 else file.read(size)
    words,realOffset=[],rOff+offset
    for i in range(0,len(data),4):
        word=int.from_bytes(data[i:i+4],"little",signed=False).to_bytes(4,"big",signed=False)
        aux=(f"{realOffset:#0{10}X}".lower())[2:]+": "
        realOffset+=4
        if noCodeZone(realOffset,type):
            value=int.from_bytes(word,"big",signed=False)
            words.append(aux+(f"{value:#0{10}X}".lower())[2:]+"\n")
        else:
            words.append(aux+getInstType(word,realOffset)+"\n")
    return words
def getInstType(word,realOffset):
    value=(int.from_bytes(word,"big",signed=False)&0xfc000000)>>26
    if value==0:
        return getZeroType(word)
    elif value==1:
        return getOneType(word,realOffset)
    elif value==2:
        return "j "+getI26(word)
    elif value==3:
        return "jal "+getI26(word)
    elif value==4:
        return "beq "+getS(word)+","+getT(word)+","+getI16(word,realOffset)
    elif value==5:
        return "bne "+getS(word)+","+getT(word)+","+getI16(word,realOffset)
    elif value==6:
        return "blez "+getS(word)+","+getI16(word,realOffset)
    elif value==7:
        return "bgtz "+getS(word)+","+getI16(word,realOffset)
    elif value==8:
        return "addi "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==9:
        return "addiu "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==10:
        return "slti "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==11:
        return "sltiu "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==12:
        return "andi "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==13:
        return "ori "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==14:
        return "xori "+getT(word)+","+getS(word)+","+getI16(word,0)
    elif value==15:
        return "lui "+getT(word)+","+getI16(word,0)
    elif value==20:
        return "beql "+getS(word)+","+getT(word)+","+getI16(word,realOffset)
    elif value==21:
        return "bnel "+getS(word)+","+getT(word)+","+getI16(word,realOffset)
    elif value==22:
        return "blezl "+getS(word)+","+getI16(word,realOffset)
    elif value==23:
        return "bgtzl "+getS(word)+","+getI16(word,realOffset)
    elif value==31:
        value = (int.from_bytes(word,"big",signed=False) & 0x0000003f)
        if value==0:
            return "ext "+getT(word)+","+getS(word)+","+getH(word)+","+getH2p1(word)
        elif value==4:
            return "ins "+getT(word)+","+getS(word)+","+getH(word)+","+getH2p1(word)
        elif value==32:
            value = (int.from_bytes(word,"big",signed=False) & 0x000007c0)>>6
            if value==2:
                return "wsbh "+getD(word)+","+getT(word)
            elif value==16:
                return "seb "+getD(word)+","+getT(word)
            elif value==24:
                return "seh "+getD(word)+","+getT(word)
            else:
                value = int.from_bytes(word,"big",signed=False)
                return (f"{value:#0{10}X}".lower())[2:]
        else:
            value = int.from_bytes(word,"big",signed=False)
            return (f"{value:#0{10}X}".lower())[2:]
    elif value==32:
        return "lb "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==33:
        return "lh "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==34:
        return "lwl "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==35:
        return "lw "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==36:
        return "lbu "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==37:
        return "lhu "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==38:
        return "lwr "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==40:
        return "sb "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==41:
        return "sh "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==42:
        return "swl "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==43:
        return "sw "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==46:
        return "swr "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==48:
        return "ll "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    elif value==56:
        return "sc "+getT(word)+","+getI16(word,0)+"("+getS(word)+")"
    else:
        value = int.from_bytes(word,"big",signed=False)
        return (f"{value:#0{10}X}".lower())[2:]
def getZeroType(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x0000003f)
    if value==0:
        if (int.from_bytes(word,"big",signed=False)==0):
            return "noop"
        else:
            return "sll "+getD(word)+","+getT(word)+","+getH(word)
    elif value==2:
        return "srl "+getD(word)+","+getT(word)+","+getH(word)
    elif value==3:
        return "sra "+getD(word)+","+getT(word)+","+getH(word)
    elif value==4:
        return "sllv "+getD(word)+","+getT(word)+","+getS(word)
    elif value==6:
        return "srlv "+getD(word)+","+getT(word)+","+getS(word)
    elif value==7:
        return "srav "+getD(word)+","+getT(word)+","+getS(word)
    elif value==8:
        return "jr "+getS(word)
    elif value==9:
        return "jalr "+getD(word)+","+getS(word)
    elif value==10:
        return "movz "+getD(word)+","+getS(word)+","+getT(word)
    elif value==11:
        return "movn "+getD(word)+","+getS(word)+","+getT(word)
    elif value==16:
        return "mfhi "+getD(word)
    elif value==17:
        return "mthi "+getS(word)
    elif value==18:
        return "mflo "+getD(word)
    elif value==19:
        return "mtlo "+getS(word)
    elif value==22:
        return "clz "+getD(word)+","+getS(word)
    elif value==23:
        return "clo "+getD(word)+","+getS(word)
    elif value==24:
        return "mult "+getS(word)+","+getT(word)
    elif value==25:
        return "multu "+getS(word)+","+getT(word)
    elif value==26:
        return "div "+getS(word)+","+getT(word)
    elif value==27:
        return "divu "+getS(word)+","+getT(word)
    elif value==28:
        return "madd "+getS(word)+","+getT(word)
    elif value==29:
        return "maddu "+getS(word)+","+getT(word)
    elif value==30:
        return "msub "+getS(word)+","+getT(word)
    elif value==31:
        return "msubu "+getS(word)+","+getT(word)
    elif value==32:
        return "add "+getD(word)+","+getS(word)+","+getT(word)
    elif value==33:
        return "addu "+getD(word)+","+getS(word)+","+getT(word)
    elif value==34:
        return "sub "+getD(word)+","+getS(word)+","+getT(word)
    elif value==35:
        return "subu "+getD(word)+","+getS(word)+","+getT(word)
    elif value==36:
        return "and "+getD(word)+","+getS(word)+","+getT(word)
    elif value==37:
        return "or "+getD(word)+","+getS(word)+","+getT(word)
    elif value==38:
        return "xor "+getD(word)+","+getS(word)+","+getT(word)
    elif value==39:
        return "nor "+getD(word)+","+getS(word)+","+getT(word)
    elif value==42:
        return "slt "+getD(word)+","+getS(word)+","+getT(word)
    elif value==43:
        return "sltu "+getD(word)+","+getS(word)+","+getT(word)
    else:
        value = int.from_bytes(word,"big",signed=False)
        return (f"{value:#0{10}X}".lower())[2:]
def getOneType(word,realOffset):
    value = (int.from_bytes(word,"big",signed=False) & 0x001f0000) >> 16
    if value==0:
        return "bltz "+getS(word)+","+getI16(word,realOffset)
    elif value==1:
        return "bgez "+getS(word)+","+getI16(word,realOffset)
    elif value==2:
        return "bltzl "+getS(word)+","+getI16(word,realOffset)
    elif value==3:
        return "bgezl "+getS(word)+","+getI16(word,realOffset)
    elif value==16:
        return "bltzal "+getS(word)+","+getI16(word,realOffset)
    elif value==17:
        return "bgezal "+getS(word)+","+getI16(word,realOffset)
    elif value==18:
        return "bltzall "+getS(word)+","+getI16(word,realOffset)
    elif value==19:
        return "bgezall "+getS(word)+","+getI16(word,realOffset)
    else:
        value = int.from_bytes(word,"big",signed=False)
        return (f"{value:#0{10}X}".lower())[2:]
def getS(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x03e00000)>>21
    return fixReg("r"+str(value))
def getT(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x001f0000)>>16
    return fixReg("r"+str(value))
def getD(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x0000f800)>>11
    return fixReg("r"+str(value))
def getH2p1(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x0000f800)>>11
    return f"{value+1:#0{4}X}".lower()
def getH(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x000007c0)>>6
    return f"{value:#0{4}X}".lower()
def getI26(word):
    value = (int.from_bytes(word,"big",signed=False) & 0x03ffffff)*4
    return f"{value:#0{10}X}".lower()
def getI16(word,offset):
    value=(int.from_bytes(word,"big",signed=False) & 0x0000ffff)
    if offset!=0:
        value=value.to_bytes(2,"big",signed=False)
        value=int.from_bytes(value,"big",signed=True)*4+offset
        return f"{value:#0{10}X}".lower()
    return f"{value:#0{6}X}".lower()
def fixReg(reg):
    return dataMIPS.MIPSrN2Na.get(reg)
def noCodeZone(realOffset,type):
    return realOffset>0x08a4b500 if type=="psp-fft-boot" else False
