#!/usr/bin/env python

import fftwlib
from fftwlib import encodeTextFiles
from fftwlib import eventMap
import multiprocessing as mp
from functools import partial
import struct

def encodeEVT(baseFolder,inFile,type):
    if inFile[1]=="disable":
        return
    print("Encoding: "+inFile[0])
    print(" CPU cores detected: "+str(mp.cpu_count()))
    baseFolder=baseFolder+inFile[0]+"/"
    fin=baseFolder+inFile[0]+"-unt"
    fou=baseFolder+inFile[0]+".bin"
    indexs,index,currOff,count=[],[],0,0
    evtPattern=b"\x7B\x65\x76\x65\x6E\x74\x20"
    with open(fin,"rb") as file:
        data=file.read()
    while currOff!=-1:
        if len(index)==2:
            count+=1
            index.extend([currOff,count])
            indexs.append(index)
            index=[]
        index.append(currOff)
        currOff=data.find(evtPattern,currOff+1)
    index.extend([len(data),count+1])
    indexs.append(index)
    encodePool=mp.Pool()
    encodeFunc=partial(encodeEVTCore,fin,type)
    encodedStr,lgMsg=bytearray(),""
    for rets in encodePool.imap(encodeFunc,indexs):
        encodedStr+=rets[0]
        lgMsg+=rets[1]+"\n"
    with open(fou,"wb") as file:
        file.write(encodedStr)
    print(lgMsg+" Done.")
    return
def encodeEVTCore(fin,type,index):
    ilines,tlines=[],[]
    with open(fin,encoding="utf8",errors="ignore") as file:
        file.seek(index[0])
        file.readline()
        while file.tell()<index[1]:
            ilines.append(file.readline())
        file.readline()
        while file.tell()<index[2]:
            tlines.append(file.readline())
    encodedStr=encodeEVTInstructions(ilines)
    pack,extraMsg=[encodedStr,[],bytearray(),[0,0,0],[]],""
    for line in tlines:
        pack=encodeTextFiles.encodeTextCore(line,False,"",pack)[0]
    encodedStr=pack[0]
    if len(encodedStr)>fftwlib.EVTsize[type]:
        extraMsg="\n  Too big: "+str(len(encodedStr)-fftwlib.EVTsize[type])
        encodedStr=encodedStr[:fftwlib.EVTsize[type]]
    return encodedStr," Event: "+str(index[3])+extraMsg
def encodeEVTInstructions(ilines):
    encodedStr=b"\x00"*4
    for line in ilines:
        lparts=line.split("(")
        inst=eventMap.EVTNa2Va.get(lparts[0])
        encodedStr+=inst[0].to_bytes(1,"big",signed=False)
        args=lparts[1].split(")")[0].split(",")
        if len(args)>1 or args[0]:
            for i in range(0, len(args)):
                args[i]=int(args[i],16 if "0x" in args[i] else 10)
            encodedStr+=struct.pack(inst[1],*args)
    return len(encodedStr).to_bytes(4,"little",signed=False)+encodedStr[4:]
