#!/usr/bin/env python
from fftwlib import dataMIPS

def encodeMIPS(fin,fou):
    words=addLittleEndian(readWords(fin))
    with open(fou,"wb") as file:
        for word in words:
            file.write(word)
    return
def readWords(fin):
    words=[]
    with open(fin,encoding="utf8",errors="ignore") as file:
        for line_number, line in enumerate(file, start=1):
            if "#" in line:
                line=line.split("#")[0]
            lparts=line.split(": ")
            words.append(setEncode(lparts[1][:-1],lparts[0]))
    return words
def addLittleEndian(list):
    words=[]
    for word in list:
        if word<0:
            print("  Bad instruction: "+str(word))
        words.append(word.to_bytes(4,"little",signed=False))
    return words
def setEncode(word,offset):
    wparts=word.split(" ")
    if len(wparts)>1:
        args=wparts[1].split(",")
    else:
        args=["","",""]
    if word.strip()=="noop":
        return 0
    elif "sll " in word:
        return setH(args[2],setT(args[1],setD(args[0],0))) | 0
    elif "srl " in word:
        return setH(args[2],setT(args[1],setD(args[0],0))) | 2
    elif "sra " in word:
        return setH(args[2],setT(args[1],setD(args[0],0))) | 3
    elif "sllv " in word:
        return setS(args[2],setT(args[1],setD(args[0],0))) | 4
    elif "srlv " in word:
        return setS(args[2],setT(args[1],setD(args[0],0))) | 6
    elif "srav " in word:
        return setS(args[2],setT(args[1],setD(args[0],0))) | 7
    elif "jr " in word:
        return setS(args[0],0) | 8
    elif "jalr " in word:
        return setS(args[1],setD(args[0],0)) | 9
    elif "movz " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 10
    elif "movn " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 11
    elif "mfhi " in word:
        return setD(args[0], 0) | 16
    elif "mthi " in word:
        return setS(args[0], 0) | 17
    elif "mflo " in word:
        return setD(args[0],0) | 18
    elif "mtlo " in word:
        return setS(args[0],0) | 19
    elif "clz " in word:
        return setS(args[1],setD(args[0],0)) | 22
    elif "clo " in word:
        return setS(args[1],setD(args[0],0)) | 23
    elif "mult " in word:
        return setT(args[1],setS(args[0],0)) | 24
    elif "multu " in word:
        return setT(args[1],setS(args[0],0)) | 25
    elif "div " in word:
        return setT(args[1],setS(args[0],0)) | 26
    elif "divu " in word:
        return setT(args[1],setS(args[0],0)) | 27
    elif "madd " in word:
        return setT(args[1],setS(args[0],0)) | 28
    elif "maddu " in word:
        return setT(args[1],setS(args[0],0)) | 29
    elif "msub " in word:
        return setT(args[1],setS(args[0],0)) | 30
    elif "msubu " in word:
        return setT(args[1],setS(args[0],0)) | 31
    elif "add " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 32
    elif "addu " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 33
    elif "sub " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 34
    elif "subu " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 35
    elif "and " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 36
    elif "nor " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 39
    elif "xor " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 38
    elif "or " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 37
    elif "slt " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 42
    elif "sltu " in word:
        return setT(args[2],setS(args[1],setD(args[0],0))) | 43
    elif "bltz " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x00",0x04000000)))
    elif "bgez " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x01",0x04000000)))
    elif "bltzl " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x02",0x04000000)))
    elif "bgezl " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x03",0x04000000)))
    elif "bltzal " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x10",0x04000000)))
    elif "bgezal " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x11",0x04000000)))
    elif "bltzall " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x12",0x04000000)))
    elif "bgezall " in word:
        return setI16(args[1],offset,setS(args[0],setT2("0x13",0x04000000)))
    elif "j " in word:
        return setI26(args[0],0x08000000)
    elif "jal " in word:
        return setI26(args[0],0x0c000000)
    elif "beq " in word:
        return setI16(args[2],offset,setT(args[1],setS(args[0],0x10000000)))
    elif "bne " in word:
        return setI16(args[2],offset,setT(args[1],setS(args[0],0x14000000)))
    elif "blez " in word:
        return setI16(args[1],offset,setS(args[0],0x18000000))
    elif "bgtz " in word:
        return setI16(args[1],offset,setS(args[0],0x1c000000))
    elif "addi " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x20000000)))
    elif "addiu " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x24000000)))
    elif "slti " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x28000000)))
    elif "sltiu " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x2c000000)))
    elif "andi " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x30000000)))
    elif "xori " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x38000000)))
    elif "ori " in word:
        return setI16(args[2],0,setS(args[1],setT(args[0],0x34000000)))
    elif "lui " in word:
        return setI16(args[1],0,setT(args[0],0x3c000000))
    elif "beql " in word:
        return setI16(args[2],offset,setT(args[1],setS(args[0],0x50000000)))
    elif "bnel " in word:
        return setI16(args[2],offset,setT(args[1],setS(args[0],0x54000000)))
    elif "blezl " in word:
        return setI16(args[1],offset,setS(args[0],0x58000000))
    elif "bgtzl " in word:
        return setI16(args[1],offset,setS(args[0],0x5c000000))
    elif "lb " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x80000000)))
    elif "lh " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x84000000)))
    elif "lwl " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x88000000)))
    elif "lw " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x8c000000)))
    elif "lbu " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x90000000)))
    elif "lhu " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x94000000)))
    elif "lwr " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0x98000000)))
    elif "sb " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0], setT(args[0],0xa0000000)))
    elif "sh " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xa4000000)))
    elif "swl " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xa8000000)))
    elif "sw " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xac000000)))
    elif "swr " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xb8000000)))
    elif "ll " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xc0000000)))
    elif "sc " in word:
        return setI16(args[1].split('(')[0],0,
                      setS(args[1].split('(')[1].split(')')[0],setT(args[0],0xe0000000)))
    elif "ext " in word:
        return setH2p1(args[3],setH(args[2],setS(args[1],setT(args[0],0x7c000000)))) | 0
    elif "ins " in word:
        return setH2p1(args[3],setH(args[2],setS(args[1],setT(args[0],0x7c000000)))) | 4
    elif "seh " in word:
        return setH("0x18",setT(args[1],setD(args[0],0x7c000000))) | 32
    elif "seb " in word:
        return setH("0x10",setT(args[1],setD(args[0],0x7c000000))) | 32
    elif "wsbh " in word:
        return setH("0x02",setT(args[1],setD(args[0],0x7c000000))) | 32
    else:
        return int(word,16)
def setS(word,value):
    return value | (int(afixReg(word.strip()).split("r")[1])<<21)
def setT(word,value):
    return value | (int(afixReg(word.strip()).split("r")[1])<<16)
def setT2(word,value):
    return value | ((int(word.strip(),16))<<16)
def setD(word,value):
    return value | (int(afixReg(word.strip()).split("r")[1])<<11)
def setH2p1(word,value):
    return value | ((int(word.strip(),16)-1)<<11)
def setH(word,value):
    return value | (int(word.strip(),16)<<6)
def setI26(word,value):
    return value | (int(word.strip(),16)>>2)
def setI16(word,offset,value):
    word=word.strip()
    if offset!=0:
        if (int(word,16)-int(offset,16)-4) < 0:
            val=-((-(int(word,16)-int(offset,16)-4))>>2)
            val=val.to_bytes(2,"big",signed=True)
            return value|int.from_bytes(val,"big",signed=False)
        return value|((int(word,16)-int(offset,16)-4)>>2)
    return value | int(word,16)
def afixReg(reg):
    return dataMIPS.MIPSNa2rN.get(reg)
