#!/usr/bin/env python

import fftwlib
from fftwlib import charMap
from fftwlib import textFileTools
from pathlib import Path
import multiprocessing as mp
from functools import partial

Dictionary=charMap.getCharMap(direct=False)

def encodeTxtFile(fin,fou,sections,files,offset):
    indexs,index,currOff,count,outMsg=[],[offset,0],0,0,""
    secPattern=b"\x7B\x7A\x65\x72\x6F\x65\x73\x28"
    with open(fin,"rb") as file:
        data=file.read()
    while True:
        currOff=data.find(secPattern,currOff)
        if currOff==-1:
            break
        off1=data.find(b"\x2C",currOff+1)
        off2=data.find(b"\x29",currOff+1)
        size=int(data[currOff+len(secPattern):off1].decode("utf8"))
        offs=int(data[off1+1:off2].decode("utf8"))
        index.extend([off2+3,count])
        indexs.append(index)
        index=[size+offs+indexs[-1][0],off2+3]
        count+=1
        currOff+=1
    if len(indexs)>1:
        tmpFiles=[]
        encodePool=mp.Pool()
        encodeFunc=partial(encodeText,fin,sections,files)
        with open(fou,"wb") as file:
            for rets in encodePool.imap(encodeFunc,indexs):
                if rets[0]:
                    tmpFiles.extend(rets[0])
                if rets[1]:
                    for val in rets[1]:
                        file.write(val)
                file.write(rets[2])
                outMsg+=rets[3]
        files.extend(tmpFiles)
    else:
        rets=encodeText(fin,sections,files,[offset,0,len(data),0])
        with open(fou,"wb") as file:
            files=rets[0]
            if rets[1]:
                for val in rets[1]:
                    file.write(val)
            file.write(rets[2])
            outMsg+=rets[3]
    print(outMsg+" Done.")
    return files
def encodeText(fin,sections,files,index):
    offset,lines,outMsg=index[0],[],""
    with open(fin,encoding="utf8",errors="ignore") as file:
        file.seek(index[1])
        while file.tell()<index[2]:
            lines.append(file.readline())
    pack=[bytearray(),[],bytearray(),[0,0,0],files] if files else None
    newSection,count=True,index[3]
    for line in lines:
        if newSection:
            count+=1
            outMsg+=" Section: "+str(count)+"\n"
        try:
            compress=sections.index(count)>=0
        except:
            compress=False
        pack,newSection,ouMsg=encodeTextCore(line,compress,fin,pack,offset)
        outMsg+=ouMsg
    return pack[4],pack[1],pack[0],outMsg
def encodeTextCore(line,compress,fin="",pack=None,offset=0):
    if pack==None:
        encodedStr,header,dEncodedStr,dInfo,files=bytearray(),[],bytearray(),[0,0,0],[]
    else:
        encodedStr,header,dEncodedStr,dInfo,files=[pack[i] for i in (0,1,2,3,4)]
    newSection,preOffset,outMsg=False,0,""
    if (not compress):
        lparts=textFileTools.charSplitter(line)
    else:
        lparts=textFileTools.charSplitterAndCompress(line)
    for val in lparts:
        if "{0x" in val:
            val=int(val[1:-1],16)
            nbyte=2 if val>255 else 1
            encodedStr+=val.to_bytes(nbyte,"big",signed=False)
            preOffset+=nbyte
        elif "{file" in val:
            dInfo[2]=int(val.split(" ")[1][:-1])
        elif "{section" in val:
            if dInfo[2]==0:
                offset=len(encodedStr)
                newSection=newSection if not header else True
                header+=offset.to_bytes(4,"little",signed=False)
            else:
                val=(val.split(" ")[1][:-1]+"-").split("-")
                if dInfo[0]==0:
                    files.append([dInfo[2]*100+int(val[0]),len(encodedStr)+offset])
                elif dInfo[0]==1:
                    dInfo[1]=dInfo[2]*100+int(val[0])
                if val[1]=="s" or val[1]=="e":
                    dInfo[0]+=1
        elif "{zeroes(" in val:
            newSection=True
            val=val.split("(")[1][:-2].split(",")
            val=int(val[0])+int(val[1])-len(encodedStr)
            if val>=0:
                encodedStr+=b'\x00'*val
                if compress:
                    outMsg+="  Font Compression Success("+str(val)+")."+"\n"
            else:
                if compress:
                    outMsg+="  Font Compression Fail("+str(val)+")."+"\n"
                    outMsg+=" This file cannot be encoded: "+fin+"\n"
                else:
                    outMsg+="  Not enough space in file: "+fin+"\n"
                    outMsg+=" This file cannot be encoded: "+fin+"\n"
        elif "{pad(" in val:
            val=val.split("(")[1][:-2].split(",")
            encodedStr+=bytes([min(int(val[0]),255)])*(int(val[1])-preOffset)
            preOffset=int(val[1])
        else:
            val=Dictionary.get(val)
            preOffset+=1
            if val>255:
                val=(val+0xcf00).to_bytes(2,"big",signed=False)
                preOffset+=1
            else:
                val=val.to_bytes(1,"big",signed=False)
            if dInfo[0]==2:
                dEncodedStr+=val
            elif dInfo[0]==4:
                files.append([dInfo[1],len(encodedStr)+offset])
                encodedStr+=dEncodedStr
                preOffset+=len(dEncodedStr)
                dEncodedStr=bytearray()
                dInfo[0]=0
            encodedStr+=val
    return [encodedStr,header,dEncodedStr,dInfo,files],newSection,outMsg
def encodeTableFile(fou,files):
    files.sort()
    with open(fou,"r+b") as file:
        file.seek(0)
        for data in files:
            file.write(data[1].to_bytes(4,"little",signed=False))
    return
def encodeTextFiles(baseFolder,files):
    print("CPU cores detected: "+str(mp.cpu_count()))
    for ifile in range(0,len(files)):
        if files[ifile][1]=="disable":
            continue
        print("Encoding: "+files[ifile][0])
        bFile=baseFolder+files[ifile][0]+"/"+files[ifile][0]
        sfiles=files[ifile][2]
        absOff=[]
        if len(files[ifile])>4:
            for offs in files[ifile][4]:
                absOff.append(offs)
        else:
            for i in range(0,len(files[ifile][2])):
                if files[ifile][2][i][0]==-1:
                    absOff.append(-1)
                else:
                    absOff.append(0)
        if ifile==0:
            textFileTools.rebuildSpellMes(bFile+"-0")
        elif ifile==1:
            textFileTools.rebuildSNPLMes(bFile+"-0")
        ifiles=[]
        for isec in range(0,len(sfiles)):
            if absOff[isec]==-1:
                continue
            offs=absOff[isec]+fftwlib.PSPramOffset
            sbFile=bFile+"-"+str(isec)
            if (Path(sbFile+"-utt").is_file()):
                print(" File: "+sbFile)
                ifiles=encodeTxtFile(sbFile+"-utt",sbFile+".bin",sfiles[isec],ifiles,offs)
            else:
                print(" File missing: "+sbFile)
        if ifiles:
            encodeTableFile(bFile+"-0.bin",ifiles)
    return
